from deciphon_snap.interval import PyInterval
from hmmer_tables.domtbl import DomTBLCoord

__all__ = ["AminoInterval", "make_amino_interval"]


class AminoInterval(PyInterval):
    ...


def make_amino_interval(x: DomTBLCoord):
    if isinstance(x, DomTBLCoord):
        return AminoInterval(start=x.interval.start, stop=x.interval.stop)
    assert False



from typing import List

from pydantic import BaseModel, RootModel

from deciphon_snap.hit import Hit, HitList
from deciphon_snap.hmmer import H3Result
from deciphon_snap.match import LazyMatchList
from deciphon_snap.query_interval import QueryIntervalBuilder

__all__ = ["Prod"]


class Prod(BaseModel):
    id: int
    seq_id: int
    profile: str
    abc: str
    alt: float
    null: float
    evalue: float
    match_list: LazyMatchList
    h3result: H3Result | None = None

    @property
    def hits(self) -> list[Hit]:
        qibuilder = QueryIntervalBuilder(self.match_list)
        hits = []
        for hit in HitList.make(self.match_list):
            hit.interval = qibuilder.make(hit.match_list_interval)
            hit.match_list = self.match_list.evaluate()
            hits.append(hit)
        return hits

    @property
    def hmmer(self):
        assert self.h3result is not None
        return self.h3result

    @property
    def query(self):
        return self.match_list.query

    @property
    def codon(self):
        return self.match_list.codon

    @property
    def amino(self):
        return self.match_list.amino


class ProdList(RootModel):
    root: List[Prod]

    def __len__(self):
        return len(self.root)

    def __getitem__(self, i) -> Prod:
        return self.root[i]

    def __iter__(self):
        return iter(self.root)



from fsspec.implementations.cached import WholeFileCacheFileSystem
from fsspec.implementations.zip import ZipFileSystem

from deciphon_snap.path_like import PathLike
from deciphon_snap.snap_file import SnapFile

__all__ = ["read_snap"]


def read_snap(filename: PathLike):
    return SnapFile(WholeFileCacheFileSystem(fs=ZipFileSystem(filename)))



from deciphon_snap.interval import PyInterval
from deciphon_snap.match import MatchList, MatchListInterval

__all__ = ["QueryInterval", "QueryIntervalBuilder"]


class QueryInterval(PyInterval):
    ...


class QueryIntervalBuilder:
    def __init__(self, match_list: MatchList):
        self._offset = []
        offset = 0
        for x in match_list:
            self._offset.append(offset)
            offset += len(x.query)
        self._offset.append(offset)

    def make(self, match_list_interval: MatchListInterval) -> QueryInterval:
        i = match_list_interval
        start = self._offset[i.pyinterval.start]
        stop = self._offset[i.pyinterval.stop]
        return QueryInterval(start=start, stop=stop)



from __future__ import annotations

from abc import ABC, abstractmethod

from pydantic import BaseModel, model_validator
from pydantic.types import Any, conint

__all__ = ["Interval", "PyInterval", "RInterval"]


class Interval(BaseModel, ABC):
    @property
    @abstractmethod
    def slice(self) -> slice:
        ...

    @property
    @abstractmethod
    def pyinterval(self) -> PyInterval:
        ...


class PyInterval(Interval):
    """
    Python interval.

    A Python interval is an 0-start, half-open interval. It means that:
    - The elements of a sequence have the coordinates `0, 1, 2, ...`.
    - An interval `PyInterval(start, stop)` is defined by the coordinates
      `start, start+1, ..., stop-2, stop-1`.

    Attributes
    ----------
    start
        Start of interval. Valid values are `0, 1, ..., stop`.
    stop
        End of interval. Valid values are `start, start+1, ...`.
    """

    start: conint(ge=0)
    stop: conint(ge=0)

    @model_validator(mode="before")
    def pre_root(cls, values: dict[str, Any]) -> dict[str, Any]:
        assert values["start"] <= values["stop"]
        return values

    @property
    def rinterval(self) -> RInterval:
        return RInterval(start=self.start + 1, stop=self.stop)

    @property
    def pyinterval(self) -> PyInterval:
        return self

    @property
    def slice(self) -> slice:
        return slice(self.start, self.stop)

    def offset(self, offset: int) -> PyInterval:
        return PyInterval(start=self.start + offset, stop=self.stop + offset)

    def __str__(self):
        return repr(self)


class RInterval(Interval):
    """
    R interval.

    An R interval is an 1-start, fully-closed. It means that:
    - The elements of a sequence have the coordinates `1, 2, 3, ...`.
    - An interval `RInterval(start, stop)` is defined by the coordinates
      `start, start+1, ..., stop-1, stop`.

    Attributes
    ----------
    start
        Start of interval. Valid values are `1, 2, ..., stop`.
    stop
        End of interval. Valid values are `start, start+1, ...`.
    """

    start: conint(gt=0)
    stop: conint(gt=0)

    @model_validator(mode="before")
    def pre_root(cls, values: dict[str, Any]) -> dict[str, Any]:
        assert values["start"] <= values["stop"]
        return values

    @property
    def pyinterval(self) -> PyInterval:
        return PyInterval(start=self.start - 1, stop=self.stop)

    @property
    def slice(self) -> slice:
        return self.pyinterval.slice

    def offset(self, offset: int) -> RInterval:
        return RInterval(start=self.start + offset, stop=self.stop + offset)

    def __str__(self):
        return repr(self)



from __future__ import annotations

from functools import lru_cache
from typing import List

from pydantic import BaseModel, ConfigDict, RootModel

from deciphon_snap.amino import AminoInterval
from deciphon_snap.interval import PyInterval

__all__ = ["Match", "MatchList", "LazyMatchList", "MatchListInterval"]


class Match(BaseModel):
    query: str
    state: str
    codon: str
    amino: str
    _position: int | None

    @classmethod
    def from_string(cls, x: str):
        y = x.split(",", 3)
        return cls(query=y[0], state=y[1], codon=y[2], amino=y[3])

    @property
    def position(self):
        assert self._position is not None
        return self._position

    @position.setter
    def position(self, x: int):
        self._position = x

    def __str__(self):
        query = self.query if len(self.query) > 0 else "∅"
        state = self.state
        codon = self.codon if len(self.codon) > 0 else "∅"
        amino = self.amino if len(self.amino) > 0 else "∅"
        return f"({query},{state},{codon},{amino})"


class MatchList(RootModel):
    root: List[Match]

    @classmethod
    def from_string(cls, x: str):
        return cls.model_validate([Match.from_string(i) for i in x.split(";")])

    def __len__(self):
        return len(self.root)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return MatchList.model_validate(self.root[i])
        match = self.root[i]
        assert isinstance(match, Match)
        return match

    def __iter__(self):
        return iter(self.root)

    def __str__(self):
        return " ".join(str(i) for i in self.root)

    @property
    def query(self):
        return "".join(x.query for x in iter(self))

    @property
    def codon(self):
        return "".join(x.codon for x in iter(self))

    @property
    def amino(self):
        return "".join(x.amino for x in iter(self))


class MatchListInterval(PyInterval):
    ...


class MatchListIntervalBuilder:
    def __init__(self, match_list: MatchList):
        self._amino_map = [i for i, x in enumerate(match_list) if len(x.amino) > 0]

    def make_from_amino_interval(
        self, amino_interval: AminoInterval
    ) -> MatchListInterval:
        i = amino_interval
        x = self._amino_map[i.slice]
        return MatchListInterval(start=x[0], stop=x[-1] + 1)


class LazyMatchList(BaseModel):
    raw: str
    model_config = ConfigDict(frozen=True)

    @lru_cache(maxsize=1)
    def evaluate(self):
        return MatchList.from_string(self.raw)

    def __len__(self):
        return len(self.evaluate())

    def __getitem__(self, i):
        return self.evaluate()[i]

    def __iter__(self):
        return iter(self.evaluate())

    def __str__(self):
        return str(self.evaluate())

    def __repr__(self):
        return repr(self.evaluate())

    @property
    def query(self):
        return self.evaluate().query

    @property
    def codon(self):
        return self.evaluate().codon

    @property
    def amino(self):
        return self.evaluate().amino



from __future__ import annotations

import csv
from typing import List

import prettytable as pt
from h3result.read_h3result import read_h3result

from deciphon_snap.hmmer import H3Result
from deciphon_snap.match import LazyMatchList
from deciphon_snap.prod import Prod
from deciphon_snap.prod import ProdList
from deciphon_snap.shorten import shorten
from deciphon_snap.stringify import stringify

__all__ = ["SnapFile"]

csv.field_size_limit(8388608)


class SnapFile:
    def __init__(self, filesystem):
        fs = filesystem

        files = fs.ls("/", detail=False)
        assert len(files) == 1

        root_dir = files[0].rstrip("/")
        assert fs.isdir(root_dir)
        prod_file = f"{root_dir}/products.tsv"

        hmmer_dir = f"{root_dir}/hmmer"
        assert fs.isdir(hmmer_dir)

        with fs.open(prod_file, "rb") as file:
            prods: List[Prod] = []
            reader = csv.DictReader((stringify(x) for x in file), delimiter="\t")
            for idx, row in enumerate(reader):
                seq_id = int(row["seq_id"])
                profile = str(row["profile"])
                with fs.open(f"{hmmer_dir}/{seq_id}/{profile}.h3r", "rb") as f2:
                    h3r = H3Result(raw=read_h3result(fileno=f2.fileno()))
                prods.append(
                    Prod(
                        id=idx,
                        seq_id=seq_id,
                        profile=profile,
                        abc=str(row["abc"]),
                        alt=float(row["alt"]),
                        null=float(row["null"]),
                        evalue=float(row["evalue"]),
                        match_list=LazyMatchList(raw=str(row["match"])),
                        h3result=h3r,
                    )
                )
            self._prods = ProdList.model_validate(prods)

    @property
    def products(self):
        return self._prods

    def __str__(self):
        fields = Prod.__fields__

        num_fields = len(fields)
        prods = [[getattr(x, i) for i in fields] for x in self.products]
        num_products = len(prods)
        if num_products >= 10:
            prods = prods[:4] + [["…"] * num_fields] + prods[-4:]

        x = pt.PrettyTable()
        x.set_style(pt.SINGLE_BORDER)
        x.field_names = fields
        x.align = "l"
        for prod in prods:
            x.add_row([shorten(i) for i in prod])

        header = f"shape: ({num_products}, {num_fields})"
        return header + "\n" + x.get_string()



from deciphon_snap.match import MatchList, LazyMatchList
from deciphon_snap.hmmer import H3Result

__all__ = ["shorten"]


def shorten(
    x: str | int | float | LazyMatchList | MatchList | H3Result, size: int = 32
):
    if isinstance(x, float):
        return f"{x:.9g}"
    if isinstance(x, int):
        return str(x)
    if isinstance(x, LazyMatchList):
        x = x.evaluate()
    if isinstance(x, MatchList):
        x = str(x)
    if isinstance(x, H3Result):
        x = str(x)
    assert isinstance(x, str)
    return x[:size] + "…" if len(x) > size else x
